class MultipartFormDataEntry:
    def __init__(self, data: bytes, content_type: str, name: str, filename: str = None, headers: dict = None):
        self.data = data
        self.content_type = content_type
        self.name = name
        self.filename = filename
        self.headers = headers if headers is not None else {}

    @classmethod
    def from_base(cls, body: bytes, headers: dict):
        assert "content-disposition" in headers, "Content-Disposition is expected"
        content_disp = list(i.strip() for i in headers['content-disposition'].split(";"))
        assert content_disp[0] == "form-data", "Content-Disposition should be form-data"
        cd_attributes = dict(x.strip().split("=") for x in content_disp[1:])
        assert "name" in cd_attributes, "Content-Disposition should have a name attribute"

        name = cd_attributes["name"]
        if name.startswith("\"") and name.endswith("\""):
            name = name[1:-1]

        filename = cd_attributes.get("filename")
        if filename is not None and filename.startswith("\"") and filename.endswith("\""):
            filename = filename[1:-1]

        return cls(body, headers.get("content-type"), name, filename, headers)
    
    def __repr__(self):
        return f"{self.__class__.__name__}({repr(self.name)}, {repr(self.data)})"

def parse_multipart_formdata(headers: dict, body: bytes):
    assert "content-type" in headers, "Content-Type is expected"
    assert headers.get("content-type").split(";")[0].strip().lower() == "multipart/form-data", "Content-Type is not multipart/form-data"
    assert int(headers.get("content-length")) == len(body), "body size is different from content-length"

    ct_attributes = dict(x.strip().split("=") for x in headers.get("content-type").split(";")[1:])
    assert "boundary" in ct_attributes, "multipart/form-data should have boundary attribute in content-type"
    boundary_base = ct_attributes['boundary']
    boundary_start = ("--" + boundary_base).encode()

    entries = body.split(boundary_start)
    print(entries)
    assert entries[0] == b"" and entries[-1] == b"--\r\n", "invalid form data"
    del entries[0]
    del entries[-1]

    entries = [entry[2:-2] for entry in entries]
    entries_with_headers = []
    for entry in entries:
        entry_headers = entry.split(b"\r\n\r\n")[0]
        entry_content = b"\r\n\r\n".join(entry.split(b"\r\n\r\n")[1:])
        entry_headers = (tuple(map(lambda x: x.decode().strip(), header.split(b":", 1))) for header in entry_headers.split(b"\r\n"))
        entry_headers = dict((x.lower(), y) for x, y in entry_headers)

        entries_with_headers.append((entry_content, entry_headers))
    
    return [MultipartFormDataEntry.from_base(entry[0], entry[1]) for entry in entries_with_headers]


from ..methods import Methods
from ..response import BinaryResponse, Response
from ..request import Request
from ..rhtml import Tag
from ..route_match import match_route_with_method
from .base import BaseConnector
import json


class HTTPConnector(BaseConnector):
    def __init__(self):
        self.routes = {}

    def route(self, route, methods=None):
        def inner(fn):
            for method in methods or [Methods.GET]:
                self.routes[route, method] = fn
            return fn
        return inner
    
    def get(self, route, other_methods=None):
        return self.route(route, [Methods.GET] + (other_methods or []))
    def post(self, route, other_methods=None):
        return self.route(route, [Methods.POST] + (other_methods or []))
    def head(self, route, other_methods=None):
        return self.route(route, [Methods.HEAD] + (other_methods or []))
    def options(self, route, other_methods=None):
        return self.route(route, [Methods.OPTIONS] + (other_methods or []))
    def trace(self, route, other_methods=None):
        return self.route(route, [Methods.TRACE] + (other_methods or []))
    def put(self, route, other_methods=None):
        return self.route(route, [Methods.PUT] + (other_methods or []))
    def delete(self, route, other_methods=None):
        return self.route(route, [Methods.DELETE] + (other_methods or []))
    def patch(self, route, other_methods=None):
        return self.route(route, [Methods.PATCH] + (other_methods or []))
    
    async def process_scope(self, scope, receive, send, reugin):
        if scope['type'] != 'http':
            return False
        
        req = Request()
        req.path = scope['path']
        req.method = scope['method']
        req.params = dict(qc.split("=") for qc in scope['query_string'].decode().split("&")) if len(scope['query_string'].decode().strip()) > 0 else {}
        req.headers = dict(map(lambda x: (y.decode() for y in x), scope['headers']))
        req.addr = scope['client']
        req._asgi_scope = scope

        if (route := match_route_with_method(req.path, Methods(scope['method'].upper()), self.routes))[0] != None:
            body = b''
            while True:
                recvscope = await receive()
                assert recvscope['type'] == 'http.request'

                body += recvscope['body']
                # bugfix: body length should be checked with <=
                assert len(body) <= reugin.max_request_body, "Max body length exceeded" 
                if recvscope['more_body'] == False:
                    break
            req.body = body # removed .decode() - body may be binary if, for example, form with binary file is submitted
            resp: Response = await route[0](req, *route[1])
            if not isinstance(resp, Response):
                if isinstance(resp, Tag):
                    resp = Response(200, "text/html", resp.render())
                elif isinstance(resp, dict):
                    resp = Response(200, "application/json", json.dumps(resp))
                else:
                    assert False, f"Expected Tag or Response, got {type(resp)}"
            await send({
                'type': 'http.response.start',
                'status': resp.code,
                'headers': [
                    [b'Content-Type', resp.content_type.encode()],
                    *[[x.encode(), str(y).encode()] for x, y in resp.headers.items()]
                ],
            })
            await send({
                'type': 'http.response.body',
                'body': resp.body.encode() if not isinstance(resp, BinaryResponse) else resp.body,
            })
            return True # finished
        else:
            return False # don't send 404 - this will be handled by root server (or further connectors)


from .base import BaseConnector
from ..response import JSONResponse
import json
import urllib
import logging

class RemoteCall:
    def __init__(self):
        self.path = None
        self.body = None
        self.method = None
        self.params = None
        self.headers = None
        self.addr = None
        self._asgi_scope = None

class RPCConnector(BaseConnector):
    def __init__(self, rpc_scope):
        self.rpc_scope = rpc_scope
        self.rpcs = {}
    
    def rpc(self, rpc_name=None):
        def inner(fn):
            nonlocal rpc_name
            if rpc_name is None:
                rpc_name = fn.__code__.co_name
            assert rpc_name not in self.rpcs, "RPC Call with this name exists already"
            self.rpcs[rpc_name] = fn
            return fn
        return inner
    
    async def process_scope(self, scope, receive, send, reugin):
        if scope['type'] != 'http':
            return False

        try:
            req = RemoteCall()
            req.path = scope['path']
            req.method = scope['method']
            req.params = dict(qc.split("=") for qc in scope['query_string'].decode().split("&")) if len(scope['query_string'].decode().strip()) > 0 else {}
            req.headers = dict(map(lambda x: (y.decode() for y in x), scope['headers']))
            req.addr = scope['client']
            req._asgi_scope = scope

            if not req.path.startswith(f"/_reuginpowered_/rpc/{self.rpc_scope}/"):
                return False
            
            for rpc_name, rpc in self.rpcs.items():
                if req.path == f"/_reuginpowered_/rpc/{self.rpc_scope}/{rpc_name}":
                    body = b''
                    while True:
                        recvscope = await receive()
                        assert recvscope['type'] == 'http.request'

                        body += recvscope['body']
                        if recvscope['more_body'] == False:
                            break
                        assert len(body) >= reugin.max_request_body and reugin.max_request_body >= 0, "Max body length exceeded"
                    req.body = body.decode()
                    try:
                        await JSONResponse(200, json.dumps(await rpc(req, *json.loads(urllib.parse.unquote(req.params.get("args", '%5B%5D')))))).send(send)
                    except TypeError:
                        await JSONResponse(400, {"status": "err", "reason": "wrong types"}).send(send)
                    except json.JSONDecodeError:
                        await JSONResponse(400, {"status": "err", "reason": "wrong json"}).send(send)
                    except Exception as e:
                        logging.exception(e)
                        await JSONResponse(500, {"status": "err", "reason": "internal server error"}).send(send)
                    return True

            # await JSONResponse(404, {"status": "err", "reason": "rpc not found"}).send(send)
            return False # don't send 404 - this will be handled by root server (or further connectors)
        except Exception as e:
            raise RuntimeError(e) # don't send 500 - this will be handled by root server (or errorhooks)


from .multipartformdata import parse_multipart_formdata as _pmfd

class Request:
    def __init__(self):
        self.path = "/"
        self.body = ""
        self.method = ""
        self.params = {}
        self.addr = None
        self.headers = {}
        self._asgi_scope = {}
    
    def get_body_json(self):
        assert "content-type" in self.headers, "Request has no Content-Type header"
        assert self.headers.get("content-type").split(";")[0].strip().lower() == "application/json", "Content-Type is not JSON"

        import json
        return json.loads(self.body.decode())
    
    def get_body_multipart_formdata(self):
        return _pmfd(self.headers, self.body)


class EmptyAttr:
    pass

class Property:
    def __init__(self, name, value):
        self.name = name
        self.value = value

class PClass(Property):
    def __init__(self, value):
        self.name = "class"
        self.value = value



from .misc import *
import html

class Tag:
    tag = "NONE"

    def __init__(self, *body, **attributes):
        self.rdbody = body
        self.rdattributes = attributes
        for attr, value in attributes.items():
            setattr(self, attr, value)
    
    def render_attribute(self, attr):
        if isinstance(attr, Property):
            attr = (attr.name, attr.value)
        name = attr[0]
        value = attr[1]
        if isinstance(value, (str, int, float)):
            value = str(value)
        elif isinstance(value, EmptyAttr):
            return f'{name}'
        elif hasattr(value, "_reugin_feature_rpc_call"):
            raise NotImplementedError("RPC calls as attribute callbacks are not implemented yet")
            value = f'alert(&quot;RPC Calls are WIP&quot;)'
        elif hasattr(value, "_reugin_unjustpython_jsname"):
            value = value._reugin_unjustpython_jsname + "();"
        else:
            raise NotImplementedError(f"Cannot serialize attribute of type {type(value)}")
        
        return f'{name}="{str(value).replace("\"","&quot;")}"'

    def render_body(self, body):
        def render_item(item):
            if isinstance(item, Tag):
                return item.render()
            elif isinstance(item, (HeadInjection, Property)):
                return ""
            else:
                return html.escape(str(item))
        return ' '.join(map(render_item, body))

    def render(self):
        #print(self.tag, list(self.rdattributes.items()) + [x for x in self.rdbody if isinstance(x, Property)])
        return f"<{self.tag}{(' ' + ' '.join(map(lambda x: self.render_attribute(x), list(self.rdattributes.items()) + [x for x in self.rdbody if isinstance(x, Property)])))}>{self.render_body(self.rdbody)}</{self.tag}>"

    def __add__(self, other):
        assert isinstance(other, (Tag, HeadInjection, Property, str)), f"Expected Tag, got {other}"
        if isinstance(self, bind):
            return bind(*self.rdbody, other)
        return bind(self, other)

class Document(Tag):
    def __init__(self, body, head=None):
        self.body = body
        self.head = head or []
    
    def head_append(self, tag):
        assert isinstance(tag, Tag), f"Expected Tag, got {tag}"
        self.head.append(tag)
        return self

    def __add__(self, other):
        assert isinstance(other, (Tag, HeadInjection, Property, str))
        return Document(self.body + [other], self.head)
    
    def fetch_injections(self, body):
        injections = {}
        for tag in body:
            if isinstance(tag, HeadInjection):
                injections[tag.name] = tag.tag
                continue
            if not isinstance(tag, Tag):
                continue
            injections.update(self.fetch_injections(tag.rdbody))
            injections.update({x.name: x.tag for x in tag.rdbody if isinstance(x, HeadInjection)})
        return injections

    def render(self):
        return f"<!DOCTYPE html><head>{self.render_body(self.head)}{"".join(map(lambda x: x.render(), self.fetch_injections(self.body).values()))}</head><body>{self.render_body(self.body)}</body>"
    

class HeadInjection:
    def __init__(self, name, tag):
        self.name = name
        self.tag = tag
class CSSInjection(HeadInjection):
    def __init__(self, name, css_link):
        self.name = name
        self.tag = link(href=css_link, rel="stylesheet")

class bind(Tag):
    def render(self):
        return self.render_body(self.rdbody)

class script(Tag):
    tag = "script"

class table(Tag):
    tag = "table"
class tr(Tag):
    tag = "tr"
class th(Tag):
    tag = "th"
class td(Tag):
    tag = "td"


class div(Tag):
    tag = "div"
class p(Tag):
    tag = "p"
class button(Tag):
    tag = "button"

class title(Tag):
    tag = "title"
class meta(Tag):
    tag = "meta"
class link(Tag):
    tag = "link"

class a(Tag):
    tag = "a"
class img(Tag):
    tag = "img"
class iframe(Tag):
    tag = "iframe"
class input(Tag):
    tag = "input"
class b(Tag):
    tag = "b"

class IncludeFile(Tag):
    def __init__(self, filename):
        self.contents = filename
        self.rdbody = []
    def render(self):
        return open(self.contents).read()
    
class IncludeString(Tag):
    def __init__(self, data):
        self.contents = data
        self.rdbody = []
    def render(self):
        return self.contents


import re

MATCHING_TYPES = {
    'alphanumeric': r'\w+',
    'path': r'[\w/]+',
    'any': r'.+',
    'filepath': r'(?!\/)(?!.*\/\.+\/)[\/aA-zZ0-9 \.]+'
}

def check_and_extract(path, pattern):
    escaped_pattern = re.escape(pattern)
    
    for type_name, regex in MATCHING_TYPES.items():
        escaped_pattern = escaped_pattern.replace(re.escape(f"{{{type_name}}}"), f"({regex})")
    
    escaped_pattern = escaped_pattern.replace(re.escape("{}"), f"({MATCHING_TYPES['any']})")
    match = re.match(f"^{escaped_pattern}$", path)
    
    if match:
        return match.groups()
    else:
        return None

def match_route(path, route_dict):
    for route, handler in route_dict.items():
        if (groups := check_and_extract(path, route)) is not None:
            return handler, groups
    return None, None

def match_route_with_method(path, method, route_dict):    
    for route, handler in route_dict.items():
        if route[1] == method and (groups := check_and_extract(path, route[0])) is not None:
            return handler, groups
    return None, None


import logging
from .connectors.base import BaseConnector

RD_VER_TRIPLE = (0, 2, 7)
RD_VER = ".".join(map(str, RD_VER_TRIPLE))
RD_ERR_404 = f"<center><h1>404 Not Found</h1><hr><small>Reugin {RD_VER}</small></center>".encode()
RD_ERR_500 = f"<center><h1>500 Internal Server Error</h1><hr><small>Reugin {RD_VER}</small></center>".encode()

class Reugin:
    max_request_body = 256 * 1024

    def __init__(self):
        self.connectors = {}
        self.errorhooks = {}
    
    def connect(self, connector, priority=100):
        if priority not in self.connectors:
            self.connectors[priority] = []
        self.connectors[priority].append(connector)
        return connector
    
    def errorhook(self, priority=100):
        def inner(fn):
            if priority not in self.errorhooks:
                self.errorhooks[priority] = []
            self.errorhooks[priority].append(fn)
            return fn
        return inner
    
    # ASGI Application
    async def __call__(self, scope, receive, send):
        try:
            for _, handlers in sorted(self.connectors.items(), key=lambda pair: pair[0]):
                for handler in handlers:
                    if await handler.process_scope(scope, receive, send, self):
                        return
        except Exception as e:
            for _, errorhooks in sorted(self.errorhooks.items(), key=lambda pair: pair[0]):
                for errorhook in errorhooks:
                    if await errorhook(scope, receive, send, self, e):
                        return
            raise e # did not handle - raise, asgi server will handle this itself
            
        raise NotImplementedError("This route has no implementation - defaults were not applied, so this error is thrown.")

    def apply_defaults(self):
        self.lifespan_handlers = []

        class DefaultsConnector(BaseConnector):
            async def process_scope(self_dc, scope, receive, send, reugin):
                if scope['type'] == 'lifespan':
                    while True:
                        message = await receive()
                        map(lambda h: h(message), self.lifespan_handlers)

                        if message['type'] == 'lifespan.startup':
                            logging.info(f"Reugin {RD_VER} is starting up!")
                            await send({'type': 'lifespan.startup.complete'})
                        elif message['type'] == 'lifespan.shutdown':
                            logging.info(f"Shutting down!")
                            await send({'type': 'lifespan.shutdown.complete'})
                            return
                elif scope['type'] == 'http':
                    await send({
                        'type': 'http.response.start',
                        'status': 404,
                        'headers': [
                            [b'Content-Type', b'text/html']
                        ],
                    })
                    await send({
                        'type': 'http.response.body',
                        'body': RD_ERR_404
                    })
                    return True
        
        self.connect(DefaultsConnector(), priority=20000)

        @self.errorhook(200)
        async def on_500(scope, receive, send, reugin, exc):
            logging.exception(exc)
            await send({
                'type': 'http.response.start',
                'status': 500,
                'headers': [
                    [b'Content-Type', b'text/html']
                ],
            })
            await send({
                'type': 'http.response.body',
                'body': RD_ERR_500
            })
            return True


from . import rhtml, mounts, unjustpython, connectors
from .server import Reugin
from .methods import Methods as methods
from .request import Request
from .connectors.rpc import RemoteCall
from .multipartformdata import parse_multipart_formdata


import inspect
import textwrap
import ast
import warnings

from .rhtml import IncludeString

class UnjustPythonTranslator:
    def transform_binop(self, node):
        operators = {
            ast.Add: '+',
            ast.Sub: '-',
            ast.Mult: '*',
            ast.Div: '/',
            ast.FloorDiv: '//',
            ast.Mod: '%',
            ast.Pow: '**'
        }
        return f"{self.transform(node.left)} {operators[type(node.op)]} {self.transform(node.right)}"
    
    def transform_compare(self, node):
        comparison_operators = {
            ast.Eq: '===',
            ast.NotEq: '!==',
            ast.Lt: '<',
            ast.LtE: '<=',
            ast.Gt: '>',
            ast.GtE: '>='
        }
        return f"{self.transform(node.left)} {comparison_operators[type(node.ops[0])]} {self.transform(node.comparators[0])}"

    def transform(self, node):
        if isinstance(node, ast.Module):
            return ';'.join(map(self.transform, node.body))
        elif isinstance(node, ast.FunctionDef):
            args = [arg.arg for arg in node.args.args]
            args_str = ', '.join(args)
            body = ';'.join(map(self.transform, node.body))
            return f"function {node.name}({args_str}) {{ {body} }}"
        elif isinstance(node, ast.ListComp):
            generator = node.generators[0]
            return f"{generator.iter.id}.map(function({generator.target.id}) {{ return {self.transform(node.elt)}; }})"
        elif isinstance(node, ast.Return):
            return f"return {self.transform(node.value)};"
        elif isinstance(node, ast.Name):
            return node.id
        elif isinstance(node, ast.Num):
            return str(node.n)
        elif isinstance(node, ast.Str):
            return f"'{node.s}'"
        elif isinstance(node, ast.Constant):
            if isinstance(node.value, bool):
                return 'true' if node.value == True else "false"
            else:
                raise NotImplementedError(node)
        elif isinstance(node, ast.BinOp):
            return self.transform_binop(node)
        elif isinstance(node, ast.Compare):
            return self.transform_compare(node)
        elif isinstance(node, ast.Call):
            return f"{self.transform(node.func)}({', '.join(map(self.transform, node.args))})"
        elif isinstance(node, ast.Expr):
            return self.transform(node.value)
        elif isinstance(node, ast.Attribute):
            return f"{self.transform(node.value)}.{node.attr}"
        elif isinstance(node, ast.Assign):
            assert len(node.targets) == 1
            return f"{self.transform(node.targets[0])} = {self.transform(node.value)}"
        elif isinstance(node, ast.If):
            return f"if({self.transform(node.test)}) {{ {';'.join(map(self.transform, node.body))} }} else {{ { ";".join(map(self.transform, node.orelse)) } }}"
        elif isinstance(node, ast.Subscript):
            return f"{self.transform(node.value)}[{self.transform(node.slice)}]"
        elif isinstance(node, ast.For):
            assert isinstance(node.target, ast.Name)
            return f"for(const {node.target.id} of {self.transform(node.iter)}) {{ {';'.join(map(self.transform, node.body))} }}"
        else:
            raise NotImplementedError(f"UnjustPython: not implemented: {ast.dump(node)}")

    def generate_js_code(self, funcname, python_code):
        return self.transform(ast.parse(python_code))

class UnjustContext:
    def __init__(self):
        self.ctx = {}
        self.octx = ""
    
    def include_file(self, filename):
        self.octx += open(filename).read()
    def include_str(self, s):
        self.octx += s
    
    def jsdef(self, name=None):
        def inner(fn):
            nonlocal name

            if name is None:
                name = fn.__code__.co_name
            assert name not in self.ctx, "This function is already registered."
            self.ctx[name] = fn
            fn._reugin_unjustpython_jsname = name
            def inner2(*a, **kw):
                raise RuntimeError("This function is intended to be called from JavaScript context.")
            inner2._reugin_unjustpython_jsname = name
            return inner2
        return inner
    
    def build_script(self):
        b = self.octx + ""
        for fn_name, fn in self.ctx.items():
            src = inspect.getsource(fn)
            src = textwrap.dedent(src)
            b += UnjustPythonTranslator().generate_js_code(fn_name, src)
        return IncludeString("<script>/* Powered by Reugin's UnjustPython code translator -- this was automatically transpiled from Python source code to JavaScript -- Reugin Framework, UnjustPython transpiler (c) 2024 by geckonerd */"+ (b) +"</script>")

def js(function):
    warnings.warn(DeprecationWarning("Reugin's Python-to-JavaScript translation layer is deprecated. It is highly unstable and is not usable in 95% of cases and will be removed in the next major update. Use JavaScript or UnjustPython (this package)."))
    src = inspect.getsource(function)
    src = textwrap.dedent(src)
    return  UnjustPythonTranslator().generate_js_code(function.__code__.co_name, src)


from .request import Request
from .response import HTMLResponse, BinaryResponse
from .server import RD_ERR_404
import os
import mimetypes
import warnings
from pathlib import Path

def StaticContent(dir):
    # disable this warning for now - Path used should cover everything
    #warnings.warn(RuntimeWarning("Static files are experimental and may cause security issues. Windows is known to be vulnerable right now - Unix-like systems should be protected though."))
    base_dir = Path(dir).resolve()
    base_path = str(base_dir)
    async def route(rq: Request, staticrel):
        file: Path = (base_dir / staticrel).resolve()
        if not str(file).startswith(base_path):
            # got outside somehow - reject
            return HTMLResponse(404, RD_ERR_404.decode())
        if not file.is_file():
            return HTMLResponse(404, RD_ERR_404.decode())
        return BinaryResponse(200, mimetypes.guess_type(str(file))[0] or "application/octet-stream", file.open("rb").read())
    return route


class Response:
    def __init__(self, code, content_type, body, headers=None):
        self.code = code
        self.content_type = content_type
        self.body = body
        self.headers = headers or {}
    async def send(self, sender):
        await sender({
            'type': 'http.response.start',
            'status': self.code,
            'headers': [
                [b'Content-Type', self.content_type.encode()],
                *[[x.encode(), str(y).encode()] for x, y in self.headers.items()]
            ],
        })
        await sender({
            'type': 'http.response.body',
            'body': self.body.encode() if not isinstance(self, BinaryResponse) else self.body,
        })

class HTMLResponse(Response):
    def __init__(self, code, body, headers=None):
        self.code = code
        self.content_type = "text/html"
        self.body = body
        self.headers = headers or {}
class JSONResponse(Response):
    def __init__(self, code, body, headers=None):
        self.code = code
        self.content_type = "application/json"
        self.body = body
        self.headers = headers or {}

        if isinstance(self.body, dict):
            import json
            self.body = json.dumps(self.body)
class BinaryResponse(Response):
    def __init__(self, code, content_type, body, headers=None):
        self.code = code
        self.content_type = content_type
        self.body = body
        self.headers = headers or {}


from .base import BaseConnector
import asyncio
from ..route_match import match_route

class WebSocketClosedException(Exception):
    def __init__(self, wsconn, code):
        self.wsconn = wsconn
        self.code = code

class WebSocketConnection:
    def __init__(self):
        self.path = None
        self.params = None
        self.headers = None
        self.addr = None
        
        self._asgi_scope = None
        self._asgi_recv = None
        self._asgi_send = None

        self._conn_broken = False
        self._conn_broken_code = 1005

        self._ws_recv_queue = []

    async def recv(self, read_even_if_closed=False):
        if self._conn_broken and read_even_if_closed:
            raise WebSocketClosedException(self, self._conn_broken_code)
        
        while len(self._ws_recv_queue) == 0:
            if self._conn_broken:
                raise WebSocketClosedException(self, self._conn_broken_code)
            await asyncio.sleep(0)
        return self._ws_recv_queue.pop(0)
    
    async def send(self, data: str | bytes):
        if self._conn_broken:
            raise WebSocketClosedException(self, self._conn_broken_code)
        
        assert isinstance(data, (str, bytes)), "websocket expected string or bytes"
        if isinstance(data, str):
            await self._asgi_send({
                "type": "websocket.send",
                "text": data
            })
        else:
            await self._asgi_send({
                "type": "websocket.send",
                "bytes": data
            })
    
    async def close(self, code=1000, reason=None):
        await self._asgi_send({
            "type": "websocket.close",
            "code": code,
            "reason": reason
        })

    async def is_closed(self):
        return self._conn_broken

class WebSocketConnector(BaseConnector):
    def __init__(self):
        self.wsroutes = {}
    
    def connect(self, ws_route):
        def inner(fn):
            self.wsroutes[ws_route] = fn
            return fn
        return inner

    async def process_scope(self, scope, receive, send, reugin):
        if scope['type'] != 'websocket':
            return False
        
        try:
            req = WebSocketConnection()
            req.path = scope['path']
            req.params = dict(qc.split("=") for qc in scope['query_string'].decode().split("&")) if len(scope['query_string'].decode().strip()) > 0 else {}
            req.headers = dict(map(lambda x: (y.decode() for y in x), scope['headers']))
            req.addr = scope['client']
            req._asgi_scope = scope
            req._asgi_recv = receive
            req._asgi_send = send
            
            if (route := match_route(req.path))[0] != None:
                while True:
                    wsscope = await receive()
                    if wsscope['type'] == "websocket.connect":
                        await send({"type": "websocket.accept"})
                        asyncio.ensure_future(route[0](req))
                    elif wsscope['type'] == "websocket.receive":
                        if "text" in wsscope:
                            data = wsscope['text']
                        else:
                            data = wsscope['bytes']
                        req._ws_recv_queue.append(data)
                    elif wsscope['type'] == "websocket.disconnect":
                        req._conn_broken = True
                        req._conn_broken_code = wsscope['code']
                        break
                    else:
                        raise NotImplementedError(f"ASGI WebSocket event not implemented {wsscope}")
                
                return True # finished
            return False # don't send 404 - this will be handled by root server (or further connectors)
        except Exception as e:
            raise RuntimeError(e) # don't send 500 - this will be handled by root server (or errorhooks)